// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: Gaetan Compere, Jean-Francois Remacle
// -------------------------------------------------------------------

#include "AdaptInterface.h"
#include "CPUTimeManager.h"
#include "MAdTimeManager.h"
#include "NodalDataManager.h"
#include "MAdStatistics.h"
#include "ModelConstraintManager.h"
#include "MAdMessage.h"
#include "History.h"

// standard C/C++
#include <iostream>
#include <stdlib.h>
#include <fstream>

#ifdef PARALLEL
#include "mpi.h"
#include "MeshDataBaseParallelInterface.h"
#endif
using std::cerr;
using std::cout;
using std::endl;
using std::set;
using std::vector;
using std::multiset;
using std::string;

namespace MAd {

  // -------------------------------------------------------------------
  MeshAdapter::MeshAdapter(pMesh m, pSField sf): 
    mesh(m),
    eSplitOp(0), eCollapseOp(0), fCollapseOp(0), descOp(0), 
    eSwapOp(0), fSwapOp(0), vMoveOp(0), rRegionOp(0), sliverFOp(0), 
    sliverROp(0), geoTracker(0), 
    mpm(MeshParametersManagerSgl::instance()), 
    verbosity(2), outPrefix(""), debugMode(false)
  {
    SFManager = new SizeFieldManager(mesh,sf);

    MAdMsgSgl            ::instance().initialize();
    MAdTimeManagerSgl    ::instance().initialize();
    MAdStatisticsSgl     ::instance().initialize();
    CallBackManagerSgl   ::instance().initialize();
    NodalDataManagerSgl  ::instance().initialize(mesh);
    MeshQualityManagerSgl::instance().initialize(mesh,SFManager->getSizeField(),MEANRATIO);
    HistorySgl           ::instance().initialize();
    HistorySgl           ::instance().closeJournal();
    CPUTimeManagerSgl    ::instance().initialize();
    ModelConstraintManagerSgl::instance().initialize(M_model(mesh));
    MeshParametersManagerSgl::instance().initialize();

    if (mesh) buildOperators();

    setDefaultValues();
  }

  // -------------------------------------------------------------------
  MeshAdapter::~MeshAdapter()
  {
    removeOperators();
    if (geoTracker) delete geoTracker;
    if (SFManager) delete SFManager;

    MAdTimeManagerSgl    ::instance().finalize();
    MAdStatisticsSgl     ::instance().finalize();
    CallBackManagerSgl   ::instance().finalize();
    NodalDataManagerSgl  ::instance().finalize();
    MeshQualityManagerSgl::instance().finalize();
    HistorySgl           ::instance().finalize();
    CPUTimeManagerSgl    ::instance().finalize();
    ModelConstraintManagerSgl::instance().finalize();
    MeshParametersManagerSgl::instance().finalize();
    MAdMsgSgl            ::instance().finalize();
  }

  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  void MeshAdapter::setDefaultValues()
  {
    algorithm = CPS_SWP_SLV_SPT;
    maxIterationsNumber = 20;

    setEdgeLenSqBounds             ( 0.25 , 2.56 ); // [ 0.5 ^ 2 , 1.6 ^ 2 ]

    if (M_dim(mesh) == 2) {
      setNoSwapQuality             ( 0.2 );
      maxIterationsNumber = 10;    
    }
    else { //  to be checked
      setNoSwapQuality             ( 0.1 );
    }
    mpm.setSwapMinImproveRatio     ( 1.2 );

    mpm.setSliverTriBound          ( 0.01 );
    mpm.setSliverTetBound          ( 0.01 );

    // infinite loops control
    setSliverPermissionInESplit    ( false );
    setSliverPermissionInECollapse ( false );

    setCollapseOnBoundary          ( false, 0. );

#ifdef PARALLEL
    load_balance_algorithm = DEFAULT_ALGORITHM;
    dataEchangerPtr = NULL;
#endif
  }

  // -------------------------------------------------------------------
  void MeshAdapter::setEdgeLenSqBounds(double lower, double upper)
  {
    if ( sqrt(lower) > 0.5*sqrt(upper) ) {
      cout << "Warning: edge length interval is too small ("
           << sqrt(lower) << ", " << sqrt(upper) <<")\n";
    }

    mpm.setLowerLengthSqBound(lower);
    mpm.setUpperLengthSqBound(upper);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::setCollapseOnBoundary(bool accept, double tolerance) 
  {
    eCollapseOp->collapseOnBoundary(accept,tolerance);
    fCollapseOp->collapseOnBoundary(accept,tolerance);
    sliverROp  ->collapseOnBoundary(accept,tolerance);
    sliverFOp  ->collapseOnBoundary(accept,tolerance);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::setAlgorithm(algorithmDefinition algo)
  {
    algorithm = algo;
  }
  // -------------------------------------------------------------------
  void MeshAdapter::setMaxIterationsNumber(int max)
  {
    maxIterationsNumber = max;
  }

  // -------------------------------------------------------------------
  void MeshAdapter::setNoSwapQuality(double noSwapQuality)
  {
    mpm.setNoSwapQuality(noSwapQuality);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::setSwapMinImproveRatio(double ratio)
  {
    mpm.setSwapMinImproveRatio(ratio);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::setSliverQuality(double sliverQuality)
  {
    mpm.setSliverTriBound(sliverQuality);
    mpm.setSliverTetBound(sliverQuality);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::setSliverPermissionInESplit(bool perm)
  {
    mpm.setSliverPermissionInESplit(perm);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::setSliverPermissionInECollapse(bool perm)
  {
    mpm.setSliverPermissionInECollapse(perm);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::setGeoTracking(bool track)
  {
    if ( track && !geoTracker) {
      geoTracker = new geoMatcher(mesh);
      geoTracker->setHomogDirichletBC();
    }
    else if ( !track && geoTracker) { 
      delete geoTracker;
      geoTracker = NULL;
    }
  }
  // -------------------------------------------------------------------
  void MeshAdapter::setVerbosity(int _verbosity)
  {
    verbosity = _verbosity;
  }

  // -------------------------------------------------------------------
  void MeshAdapter::clearConstraints() const
  {
    DeleteConstraint(mesh);
    ModelConstraintManagerSgl::instance().clear();
  }

  // -------------------------------------------------------------------
  void MeshAdapter::setConstraint(pEntity e) const
  {
    EN_constrain(e);
  
    switch (EN_type(e)) {
    case 1:
      EN_constrain( (pEntity)E_vertex((pEdge)e,0) );
      EN_constrain( (pEntity)E_vertex((pEdge)e,1) );
      break;
    case 2:
      setConstraint( (pEntity)F_edge((pFace)e,0) );
      setConstraint( (pEntity)F_edge((pFace)e,1) );
      setConstraint( (pEntity)F_edge((pFace)e,2) );
      break;
    case 3:
      setConstraint( (pEntity)R_face((pRegion)e,0) );
      setConstraint( (pEntity)R_face((pRegion)e,1) );
      setConstraint( (pEntity)R_face((pRegion)e,2) );
      setConstraint( (pEntity)R_face((pRegion)e,3) );
      break;    
    }
  }

  // -------------------------------------------------------------------
  void MeshAdapter::setConstraint(int type, int id) const
  {
    ModelConstraintManagerSgl::instance().constrain(type,id);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::removeConstraint(int type, int id) const
  {
    ModelConstraintManagerSgl::instance().unconstrain(type,id);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::setConstraint(pGEntity e) const
  {
    ModelConstraintManagerSgl::instance().constrain(e);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::removeConstraint(pGEntity e) const
  {
    ModelConstraintManagerSgl::instance().unconstrain(e);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::removeOperators()
  {
    if (eSplitOp)    { delete eSplitOp;     eSplitOp=0;    }
    if (eCollapseOp) { delete eCollapseOp;  eCollapseOp=0; }
    if (fCollapseOp) { delete fCollapseOp;  fCollapseOp=0; }
    if (descOp)      { delete descOp;       descOp=0;      }
    if (eSwapOp)     { delete eSwapOp;      eSwapOp=0;     }
    if (fSwapOp)     { delete fSwapOp;      fSwapOp=0;     }
    if (vMoveOp)     { delete vMoveOp;      vMoveOp=0;     }
    if (rRegionOp)   { delete rRegionOp;    rRegionOp=0;   }
    if (sliverFOp)   { delete sliverFOp;    sliverFOp=0;   }
    if (sliverROp)   { delete sliverROp;    sliverROp=0;   }
  }

  // -------------------------------------------------------------------
  void MeshAdapter::buildOperators()
  {
    removeOperators();

    eSplitOp    = new edgeSplitOp         (mesh,SFManager->getSizeField());
    eCollapseOp = new edgeCollapseOp      (mesh,SFManager->getSizeField());
    fCollapseOp = new faceCollapseOp      (mesh,SFManager->getSizeField());
    descOp      = new DESCOp              (mesh,SFManager->getSizeField());
    eSwapOp     = new edgeSwapOp          (mesh,SFManager->getSizeField());
    fSwapOp     = new faceSwapOp          (mesh,SFManager->getSizeField());
    vMoveOp     = new vertexMoveOp        (mesh,SFManager->getSizeField(),false);
    rRegionOp   = new regionRemoveOp      (mesh,SFManager->getSizeField());
    sliverROp   = new sliverRegionHandler (mesh,SFManager->getSizeField());
    sliverFOp   = new sliverFaceHandler   (mesh,SFManager->getSizeField());
  }

  // -------------------------------------------------------------------
  void MeshAdapter::addCallback(CBFunction CB, void* userData, 
                                CBFunc_move CB_move, void* userData_move)
  { 
    CallBackManagerSgl::instance().registerCallBack(CB,userData);
    if (CB_move) {
      CallBackManagerSgl::instance().registerCallBackMove(CB_move,userData_move);
    }
  }

  // -------------------------------------------------------------------
  void MeshAdapter::addSizeField(pSField sf)
  {
    SFManager->addSizeField(sf);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::incrementTime(double dt) 
  {
    MAdTimeManagerSgl::instance().incrementTime(dt);
    updateSizeField();
  }

  // -------------------------------------------------------------------
  void MeshAdapter::setTime(double t) 
  {
    MAdTimeManagerSgl::instance().setTime(t);
    updateSizeField();
  }

  // -------------------------------------------------------------------
  double MeshAdapter::getTime() const {
    return MAdTimeManagerSgl::instance().getTime();
  }

  // -------------------------------------------------------------------
  void MeshAdapter::updateSizeField()
  {
    SFManager->update();
  }

  // -------------------------------------------------------------------
  void MeshAdapter::storeInitialCoordinates()
  {
    if ( getTime() != 0. ) {
      cout << "Warning: storing coordinates at time " << getTime() << endl;
    }
    NodalDataManagerSgl::instance().storeCoordinates();
  }

  // -------------------------------------------------------------------
  void MeshAdapter::removeStoredCoordinates()
  {
    NodalDataManagerSgl::instance().removeCoordinates();
  }

  // -------------------------------------------------------------------
  void MeshAdapter::registerObjects(mobileObjectSet* objs)
  {
    set<mobileObject*> allObj = objs->getObjects();
  
    set<mobileObject*>::const_iterator it = allObj.begin();
    for (; it != allObj.end(); it++) {
      set<LocalSizeField* > sizes = (*it)->getSizes();
      set<LocalSizeField* >::iterator it = sizes.begin();
      for (; it != sizes.end(); it++) {
        SFManager->addSizeField(*it);
      }
    }

    objects = objs;
  }

  // -------------------------------------------------------------------
  void MeshAdapter::registerData  (string name, 
                                   const vector<double> data) const
  {
    NodalDataManagerSgl::instance().registerData(name,data);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::registerVData (string name, 
                                   const vector<vector<double> > data) const
  {
    NodalDataManagerSgl::instance().registerVData(name,data);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::getMeshData   (string name, 
                                   vector<double> * data) const
  {
    NodalDataManagerSgl::instance().getMeshData(name,data);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::getMeshVData  (string name, 
                                   vector<vector<double> > * data) const
  {
    NodalDataManagerSgl::instance().getMeshVData(name,data);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::removeData    (string name) const
  {
    NodalDataManagerSgl::instance().removeData(name);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::removeVData   (string name) const
  {
    NodalDataManagerSgl::instance().removeVData(name);
  }

  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  // LEVEL 1 OPERATIONS
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  bool MeshAdapter::splitEdge(pEdge edge)
  {
    // set edge to split
    eSplitOp->setSplitEdge(edge);

    // find the coordinates of the new vertex (mid point)
    double xyznew[3];
    E_center(edge,xyznew);
    eSplitOp->setSplitPos(xyznew);

    // do the checks
    double worstShape;
    if ( !eSplitOp->evaluate(&worstShape) ) return false;

    // check that we do not create a worse sliver ( if asked for )
    if ( !mpm.getSliverPermissionInESplit() ) {
      if ( worstShape < mpm.getSliverTetBound() ) {
        double oriWorst;
        MeshQualityManagerSgl::instance().E_worstShape(edge, &oriWorst);
        if ( worstShape < oriWorst )    return false;
      }
    }

    // do the job
    eSplitOp->apply();
    return true;
  }

  // -------------------------------------------------------------------
  bool MeshAdapter::collapseEdge(pEdge edge)
  {
    double shapes[2] = {-1.,-1.};
    int ok[2] = {0,0};

    for (int iDir=0; iDir<2; iDir++) {

      pVertex vDel = E_vertex(edge,iDir);
      pVertex vTgt = E_vertex(edge,1-iDir);
      eCollapseOp->setCollapseEdge(edge,vDel,vTgt);

      //     eCollapseOp->setCollapseEdge(edge,iDir);

      // do the checks
      ok[iDir] = eCollapseOp->evaluate(&(shapes[iDir]));
    
      // check that we do not create a worse sliver ( if asked for )
      if ( !mpm.getSliverPermissionInECollapse() ) {
        if ( shapes[iDir] < mpm.getSliverTetBound() ) {
          double oriWorst;
          MeshQualityManagerSgl::instance().E_worstShape(edge, &oriWorst);
          if ( shapes[iDir] < oriWorst ) ok[iDir] = 0;
        }
      }
    }
  
    // find the best direction
    int best = -1; double worst = -1.;
    if ( ok[0] ) { best = 0; worst = shapes[0]; }
    if ( ok[1] && shapes[1] > worst ) {
      best = 1;
      worst = shapes[1];
    }
  
    // if possible apply an edge collapse
    if ( best >= 0 ) {
      pVertex vDel = E_vertex(edge,best);
      pVertex vTgt = E_vertex(edge,1-best);
      eCollapseOp->setCollapseEdge(edge,vDel,vTgt);
      eCollapseOp->apply();
      return true;
    }

    return false;
  }

  // -------------------------------------------------------------------
  bool MeshAdapter::collapseFace(pFace face, pEdge edge)
  {
    //   for( int ClpsOnvt = 1; ClpsOnvt >=0; ClpsOnvt-- ) {
    for( int ClpsOnvt = 0; ClpsOnvt < 2; ClpsOnvt++ ) {
      fCollapseOp->reset(face, edge, ClpsOnvt);
  
      // do the checks
      double worstShape;
      if ( !fCollapseOp->evaluate(&worstShape) ) continue;
  
      // check that we do not create a sliver
      if ( worstShape < mpm.getSliverTetBound() ) continue;
  
      // do the job
      fCollapseOp->apply();
      return true;
    }
    return false;
  }

  // -------------------------------------------------------------------
  bool MeshAdapter::DSplitCollapseEdge(pRegion pr, pEdge edge1, pEdge edge2)
  {
    descOp->setDESC(pr,edge1,edge2);

    // do the checks
    double worstShape;
    if ( !descOp->evaluate(&worstShape) ) return false;

    // check that we do not create a sliver
    if ( worstShape < mpm.getSliverTetBound() ) return false;

    // do the job
    descOp->apply();
    return true;
  }

  // -------------------------------------------------------------------
  bool MeshAdapter::swapEdge(pEdge edge)
  {
    eSwapOp->setSwapEdge(edge);

    // do the checks
    double worstShape;
    if ( !eSwapOp->evaluate(&worstShape) ) return false;

    // check that we do not create a sliver
    if ( worstShape < mpm.getSliverTetBound() ) return false;

    // do the job
    eSwapOp->apply();
    return true;
  }

  // -------------------------------------------------------------------
  bool MeshAdapter::swapFace(pFace face)
  {
    fSwapOp->setSwapFace(face);
  
    // do the checks
    double worstShape;
    if ( !fSwapOp->evaluate(&worstShape) ) return false;

    // check that we do not create a sliver
    if ( worstShape < mpm.getSliverTetBound() )  return false;

    // do the job
    fSwapOp->apply();
    return true;
  }

  // -------------------------------------------------------------------
  bool MeshAdapter::moveVertex (pVertex v, double dxyz[3])
  {
    return vMoveOp->move(v,dxyz);
  }

  // -------------------------------------------------------------------
  bool MeshAdapter::removeRegion(pRegion region)
  {
    rRegionOp->setRegion(region);
  
    // do the checks
    double worstShape;
    if ( !rRegionOp->evaluate(&worstShape) ) return false;

    // check that we do not create a sliver
    if ( worstShape < mpm.getSliverTetBound() )  return false;

    // do the job
    rRegionOp->apply();
    return true;
  }

  // -------------------------------------------------------------------
  bool MeshAdapter::putVertex (pVertex v, double xyz[3])
  {
    double oriPos[3], dxyz[3];
    V_coord(v,oriPos);
    for (int i=0; i<3; i++) dxyz[i] = xyz[i] - oriPos[i];
    return moveVertex(v,dxyz);
  }

  // -------------------------------------------------------------------
  bool MeshAdapter::moveVertices (multiset<vDisplacement,vDisplacementLess>& vDisps)
  {
    return vMoveOp->move(vDisps);
  }

  // -------------------------------------------------------------------
  // LEVEL 2 OPERATIONS
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  double MeshAdapter::LaplaceSmoothing(LaplSmooType type)
  {
    double convergenceCriterion = 1e-2;
    double L2Disp = 0.;
    LaplaceSmoothingOp* laplOp = new LaplaceSmoothingOp(mesh,SFManager->getSizeField());
    laplOp->run(type,&L2Disp);
    double L2Disp0 = L2Disp; int i = 0;
    double L2DispTot = 0.;
    while ( ( L2Disp > convergenceCriterion * L2Disp0 ) && ( L2Disp > MAdTOL ) ) {
      laplOp->run(type,&L2Disp);
      i++;
      L2DispTot += L2Disp;
    }
    if (laplOp) delete laplOp;
    return L2DispTot;
  }

  // -------------------------------------------------------------------
  int MeshAdapter::eSplitLoop() 
  {
    int nSplit = 0;
    int countE = 0; int oriNbEdges = M_numEdges(mesh);
    EIter ei = M_edgeIter(mesh);
    pEdge edge;
    while ( (countE < oriNbEdges) && ( edge = EIter_next(ei) ) ) {
      double lengthSq = SFManager->getSizeField()->SF_E_lengthSq(edge);
      if ( lengthSq > mpm.getUpperLengthSqBound() ) if (splitEdge(edge)) nSplit++;
      countE++;
    }
    EIter_delete(ei);
    return nSplit;
  }

  // -------------------------------------------------------------------
  int MeshAdapter::eCollapseLoop() 
  {
    int nCollapse = 0;
    int countE = 0; int oriNbEdges = M_numEdges(mesh);
    EIter ei = M_edgeIter(mesh);
    pEdge edge;
    while ( (countE < oriNbEdges) && ( edge = EIter_next(ei) ) ) {
      double lengthSq = SFManager->getSizeField()->SF_E_lengthSq(edge);
      if ( lengthSq < mpm.getLowerLengthSqBound() ) if (collapseEdge(edge)) nCollapse++;
      countE++;
    }
    EIter_delete(ei);
    return nCollapse;
  }

  // -------------------------------------------------------------------
  int MeshAdapter::eSplitCollapseLoop() 
  {
    int nSplit = 0;
    int nCollapse = 0;
    int countE = 0; int oriNbEdges = M_numEdges(mesh);
    EIter ei = M_edgeIter(mesh);
    pEdge edge;
    while ( (countE < oriNbEdges) && ( edge = EIter_next(ei) ) ) {
      double lengthSq = SFManager->getSizeField()->SF_E_lengthSq(edge);
      if ( lengthSq > mpm.getUpperLengthSqBound() ) if (splitEdge(edge))    nSplit++;
      if ( lengthSq < mpm.getLowerLengthSqBound() ) if (collapseEdge(edge)) nCollapse++;
      countE++;
    }
    EIter_delete(ei);
    return nSplit+nCollapse;
  }

  // -------------------------------------------------------------------
  int MeshAdapter::edgeSwapLoop()
  {
    int nb_eswap = 0;
    pEdge edge;
    EIter ei = M_edgeIter(mesh);
    int ne = M_numEdges(mesh);
    int i = 0;
    while( ( edge = EIter_next(ei) ) )
    {
      if( i > ne ) break;
      i++;
  
      if( EN_constrained((pEntity)edge) )  continue;
      if( E_numRegions(edge) > eSwapOp->getMaxNumRgns() )  continue;
  
      double worstShape;
      MeshQualityManagerSgl::instance().E_worstShape(edge, &worstShape);
  
      if( worstShape > mpm.getNoSwapQuality() ) continue;
  
      eSwapOp->setSwapEdge(edge);

      double newWorstShape;
      if( eSwapOp->evaluate(&newWorstShape) )
      {
        if( newWorstShape > mpm.getSwapMinImproveRatio()*worstShape )
        {
          if( ( eSwapOp->getMaxLenSq() < mpm.getUpperLengthSqBound() ) &&
            ( eSwapOp->getMinLenSq() > mpm.getLowerLengthSqBound() ) )
          {
            eSwapOp->apply();
            nb_eswap++;
          }
        }
      }
    }
    EIter_delete(ei);
    return nb_eswap;
  }

  // -------------------------------------------------------------------
  int MeshAdapter::faceSwapLoop() 
  {
    int nfswap = 0;
    pFace face;
    FIter fi = M_faceIter(mesh);
    int nf = M_numFaces(mesh);
    int i = 0;
    while ( ( face = FIter_next(fi) ) ) {
      if( i > nf ) break;
      i++;

      if( EN_constrained((pEntity)face) )  continue;

      double oriWorst;
      MeshQualityManagerSgl::instance().F_worstShape(face, &oriWorst);

      if ( oriWorst > mpm.getNoSwapQuality() ) continue;

      fSwapOp->setSwapFace(face);
      double newWorst;
      if( fSwapOp->evaluate(&newWorst) ) {
        if ( newWorst > mpm.getSwapMinImproveRatio()*oriWorst ) {
          fSwapOp->apply();
          nfswap++;
        }
      }
    }
    FIter_delete(fi);
    return nfswap;
  }

  // -------------------------------------------------------------------
  void MeshAdapter::removeSliverFaces(int * nbBefore, int * nbAfter)
  {
    sliverFOp->removeSliverFaces(nbBefore,nbAfter);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::removeSliverRegions(int * nbBefore, int * nbAfter)
  {
    sliverROp->removeSliverRegions(nbBefore,nbAfter);
  }

  // -------------------------------------------------------------------
  // LEVEL 3 OPERATIONS
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  int MeshAdapter::optimiseEdgeLength()
  {
    int nbMaxItOptiLen = 10;
    int nSplColl = 0;
    for (int iter=0; iter < nbMaxItOptiLen; iter++ ) {
      int nSC = eSplitCollapseLoop();
      cout << "Applied "<<nSC<<" split or collapses\n";
      if (!nSC) break;
      nSplColl += nSC;
    }
    return nSplColl;
  }

  // -------------------------------------------------------------------
  int MeshAdapter::optimiseElementShape()
  {
    int nbMaxItOptiShp = 10;
    int nESwapTot = 0, nFSwapTot = 0;
    int nESwap = 0, nFSwap = 0;
    for (int iter=0; iter < nbMaxItOptiShp; iter++ ) {
      // find ill-shaped elements around each edge
      nESwap = edgeSwapLoop();
      cout << "Applied " << nESwap<< " edge swaps"<<endl;
      if ( M_dim(mesh) == 3 ) {
        // find ill-shaped elements around each face
        nFSwap = faceSwapLoop();
        cout << "Applied " << nFSwap<< " face swaps"<<endl;
      }
    
      if (nESwap+nFSwap == 0) break;
      nESwapTot += nESwap;
      nFSwapTot += nFSwap;
    }
    return ( nESwapTot + nFSwapTot );
  }

  // -------------------------------------------------------------------
  int MeshAdapter::runOneIter()
  {
    CPUTimeManager& tm   = CPUTimeManagerSgl::instance();
    MAdStatistics&  stat = MAdStatisticsSgl ::instance();

    int numTopoChg = 0;
#ifdef PARALLEL
    { 
      double t0= tm.getTime();
      UpdateParallelConstraint(mesh);
      double dt = tm.getTime() - t0;
      if( verbosity >= 2 ) {
        cout<<endl<<"Performed constraint of parallel interface in "
            <<dt<<"seconds."<<endl;
      } 
      MPI_Barrier(MPI_COMM_WORLD);
    }
#endif 
    if ( algorithm == CPS_SWP_SLV_SPT ) {

      // Collapse short edges
      // ---------------------
      {
        double t0 = tm.getTime();
        int numCollapse = eCollapseLoop();
        double dt = tm.getTime() - t0;
        stat.addCPUECollapses(dt);
        stat.addNumECollapses(numCollapse);
        if ( verbosity >= 2 ) {
          cout << "Performed "<< numCollapse<<" edge collapses in "<<dt<<" seconds\n";
        }
        numTopoChg += numCollapse;
      }

      // Swap edges to improve elements quality
      // ---------------------------------------
      {
        double t0 = tm.getTime();
        int numSwap = edgeSwapLoop();
        double dt = tm.getTime() - t0;
        stat.addCPUESwaps(dt);
        stat.addNumESwaps(numSwap);
        if ( verbosity >= 2 ) {
          cout << "Performed "<< numSwap<<" edge swaps in "<<dt<<" seconds\n";
        }
        numTopoChg += numSwap;
      }

      // Remove slivers
      // ---------------
      if (M_dim(mesh) == 3) {
        {
          int nbBefore = 0;
          int nbAfter  = 0;
          double t0 = tm.getTime();
          removeSliverRegions(&nbBefore,&nbAfter);
          double dt = tm.getTime() - t0;
          stat.addCPURSlivers(dt);
          if ( verbosity >= 2 ) {
            cout << "Sliver regions " << nbBefore << " -> " << nbAfter << " in "<<dt<<" seconds\n";
          }
          numTopoChg += ( nbBefore - nbAfter );
        }
      }
      else {
        {
          int nbBefore = 0;
          int nbAfter  = 0;
          double t0 = tm.getTime();
          removeSliverFaces(&nbBefore,&nbAfter);
          double dt = tm.getTime() - t0;
          stat.addCPUFSlivers(dt);
          if ( verbosity >= 2 ) {
            cout << "Sliver faces " << nbBefore << " -> " << nbAfter << " in "<<dt<<" seconds\n";
          }
          numTopoChg += ( nbBefore - nbAfter );
        }
      }

      // Split long edges
      // -----------------
      {
        double t0 = tm.getTime();
        int numSplit = eSplitLoop();
        double dt = tm.getTime() - t0;
        stat.addCPUESplits(dt);
        stat.addNumESplits(numSplit);
        if ( verbosity >= 2 ) {
          cout << "Performed "<< numSplit<<" edge splits in "<<dt<<" seconds\n";
        }
        numTopoChg += numSplit;
      }

    }

    if ( algorithm == SPT_SWP_SLV_CPS ) {

      // Split long edges
      // -----------------
      {
        double t0 = tm.getTime();
        int numSplit = eSplitLoop();
        double dt = tm.getTime() - t0;
        stat.addCPUESplits(dt);
        stat.addNumESplits(numSplit);
        if ( verbosity >= 2 ) {
          cout << "Performed "<< numSplit<<" edge splits in "<<dt<<" seconds\n";
        }
        numTopoChg += numSplit;
      }

      // Swap edges to improve elements quality
      // ---------------------------------------
      {
        double t0 = tm.getTime();
        int numSwap = edgeSwapLoop();
        double dt = tm.getTime() - t0;
        stat.addCPUESwaps(dt);
        stat.addNumESwaps(numSwap);
        if ( verbosity >= 2 ) {
          cout << "Performed "<< numSwap<<" edge swaps in "<<dt<<" seconds\n";
        }
        numTopoChg += numSwap;
      }

      // Remove slivers
      // ---------------
      if (M_dim(mesh) == 3) {
        {
          int nbBefore = 0;
          int nbAfter  = 0;
          double t0 = tm.getTime();
          removeSliverRegions(&nbBefore,&nbAfter);
          double dt = tm.getTime() - t0;
          stat.addCPURSlivers(dt);
          if ( verbosity >= 2 ) {
            cout << "Sliver regions " << nbBefore << " -> " << nbAfter << " in "<<dt<<" seconds\n";
          }
          numTopoChg += ( nbBefore - nbAfter );
        }
      }
      else {
        {
          int nbBefore = 0;
          int nbAfter  = 0;
          double t0 = tm.getTime();
          removeSliverFaces(&nbBefore,&nbAfter);
          double dt = tm.getTime() - t0;
          stat.addCPUFSlivers(dt);
          if ( verbosity >= 2 ) {
            cout << "Sliver faces " << nbBefore << " -> " << nbAfter << " in "<<dt<<" seconds\n";
          }
          numTopoChg += ( nbBefore - nbAfter );
        }
      }

      // Collapse short edges
      // ---------------------
      {
        double t0 = tm.getTime();
        int numCollapse = eCollapseLoop();
        double dt = tm.getTime() - t0;
        stat.addCPUECollapses(dt);
        stat.addNumECollapses(numCollapse);
        if ( verbosity >= 2 ) {
          cout << "Performed "<< numCollapse<<" edge collapses in "<<dt<<" seconds\n";
        }
        numTopoChg += numCollapse;
      }

    }

    if ( algorithm == SLV_CPS_SWP_SPT ) {

      // Remove slivers
      // ---------------
      if (M_dim(mesh) == 3) {
        {
          int nbBefore = 0;
          int nbAfter  = 0;
          double t0 = tm.getTime();
          removeSliverRegions(&nbBefore,&nbAfter);
          double dt = tm.getTime() - t0;
          stat.addCPURSlivers(dt);
          if ( verbosity >= 2 ) {
            cout << "Sliver regions " << nbBefore << " -> " << nbAfter << " in "<<dt<<" seconds\n";
          }
          numTopoChg += ( nbBefore - nbAfter );
        }
      }
      else {
        {
          int nbBefore = 0;
          int nbAfter  = 0;
          double t0 = tm.getTime();
          removeSliverFaces(&nbBefore,&nbAfter);
          double dt = tm.getTime() - t0;
          stat.addCPUFSlivers(dt);
          if ( verbosity >= 2 ) {
            cout << "Sliver faces " << nbBefore << " -> " << nbAfter << " in "<<dt<<" seconds\n";
          }
          numTopoChg += ( nbBefore - nbAfter );
        }
      }

      // Collapse short edges
      // ---------------------
      {
        double t0 = tm.getTime();
        int numCollapse = eCollapseLoop();
        double dt = tm.getTime() - t0;
        stat.addCPUECollapses(dt);
        stat.addNumECollapses(numCollapse);
        if ( verbosity >= 2 ) {
          cout << "Performed "<< numCollapse<<" edge collapses in "<<dt<<" seconds\n";
        }
        numTopoChg += numCollapse;
      }

      // Swap edges to improve elements quality
      // ---------------------------------------
      {
        double t0 = tm.getTime();
        int numSwap = edgeSwapLoop();
        double dt = tm.getTime() - t0;
        stat.addCPUESwaps(dt);
        stat.addNumESwaps(numSwap);
        if ( verbosity >= 2 ) {
          cout << "Performed "<< numSwap<<" edge swaps in "<<dt<<" seconds\n";
        }
        numTopoChg += numSwap;
      }

      // Split long edges
      // -----------------
      {
        double t0 = tm.getTime();
        int numSplit = eSplitLoop();
        double dt = tm.getTime() - t0;
        stat.addCPUESplits(dt);
        stat.addNumESplits(numSplit);
        if ( verbosity >= 2 ) {
          cout << "Performed "<< numSplit<<" edge splits in "<<dt<<" seconds\n";
        }
        numTopoChg += numSplit;
      }

    }

    //   // Reposition the vertices
    //   // ------------------------
    //     double t0 = tm.getTime();
    //     double L2Disp = LaplaceSmoothing(OPTIMAL);
    // //     double L2Disp = LaplaceSmoothing(FAST);
    //     double dt = tm.getTime() - t0;
    //     stat.addLaplace(dt);
    //       if ( verbosity >= 2 ) {
    //     cout << "Performed a Laplace smoothing: total L2 norm of the displacements: "<<L2Disp<<" in "<<dt<<" seconds\n";
    //       }

#ifdef PARALLEL
    // Move partitions interfaces
    // ---------------------------
    { 
      MPI_Barrier(MPI_COMM_WORLD);
      double t0= tm.getTime();
      DeleteParallelConstraint(mesh);
      double dt = tm.getTime() - t0;
      if( verbosity >= 2 ) {
        cout<<endl<<"Performed unconstraint of parallel interface in "
            <<dt<<"seconds."<<endl;
      }
      if( numTopoChg !=0 ){
        t0= tm.getTime();
        switch( load_balance_algorithm  ) {      
        case DEFAULT_ALGORITHM:
          Balance2( mesh, *dataEchangerPtr );
          break;
        case METIS_ALGORITHM:
#ifdef _HAVE_PARMETIS_
          BalanceMetis2( mesh, *dataEchangerPtr );
#else
          cout<<end<<" Balancing by using PARMETIS is not allowed"<<endl;
          assert(0);
#endif
          break;
        default:
          cout<<endl<<endl<<" Unknown load balancing algorithm: "
              << load_balance_algorithm<<endl;
          assert(0);        
        }
        dt = tm.getTime() - t0;
        if( verbosity >= 2 ) {
          cout<<endl<<"Performed load balancing in "
              <<dt<<"seconds."<<endl;
        }
      }
    }
#endif

    // check the mesh (debug)
    // -----------------------
    if (debugMode) {
      std::cout << "Checking mesh validity (debug mode)\n";
      if ( !checkMesh(mesh,CHECK_ALL,1) )  abort();
    }

    return numTopoChg;
  }

  // -------------------------------------------------------------------
  void MeshAdapter::uglyTheMesh(double avgQualThresh, int maxIt)
  {
    MeshQualityManagerSgl::instance().evaluateStatistics();
    double meanQuality  = MeshQualityManagerSgl::instance().getMeanShape();
  
    int iter = 0;
    while ( (meanQuality > avgQualThresh) && (iter < maxIt) ) {

      int numOp = 0;
      int oriNumFaces = M_numFaces(mesh);
      int count = 0;
      FIter fi = M_faceIter(mesh);
      pFace pf;
      while ( ( pf = FIter_next(fi) ) && ( count < oriNumFaces ) ) {
        int flag = 0;
        for (int j=0; j<3; j++) {
          pEdge pe = F_edge(pf,j);
          for( int ClpsOnvt = 0; ClpsOnvt < 2; ClpsOnvt++ ) {
            fCollapseOp->reset(pf, pe, ClpsOnvt);
            double worstShape;
            if ( fCollapseOp->evaluate(&worstShape) ) {
              fCollapseOp->apply();
              flag = 1;
              break;
            }
          }
          if ( flag ) {
            numOp++;
            break;
          }
        }
        count++;
      }
      FIter_delete(fi);
      cout << "Num face collapses: "<<numOp<<endl;
    
      meanQuality  = MeshQualityManagerSgl::instance().getMeanShape();
      iter++;
    }
  }

  // -------------------------------------------------------------------
  //! Removes elements with negative volume and returns \ingroup adaptation
  //! number of remaining negative volumes
  int MeshAdapter::removeNegativeElements()
  {
    int nNeg = 0;
    int nRem = 0;
    RIter rIt= M_regionIter(mesh);
    while ( pRegion region = RIter_next(rIt) ) {
      if (R_volume(region) < 0.) {
        nNeg++;
        if ( removeRegion(region) ) nRem++;
      }
    }
    RIter_delete(rIt);

    cout << "Negative elements removal: "
         << nNeg << " detected, "
         << nRem << " removed\n";

    return (nNeg - nRem);
  }

  // -------------------------------------------------------------------
  // LEVEL 4 OPERATIONS
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  void MeshAdapter::run()
  {
    // --- Adaptation procedure ---

    cout << "\n +++ Starting a MDB mesh adaptation procedure +++\n";
    if ( verbosity >= 2 ) cout<<"\n";

    int maxOp = 0;
    int minOp = 100000000;

    for (int iter=1; iter <= maxIterationsNumber; iter++) {
    
      if ( verbosity >= 1 ) {
        cout<<"   --- Mesh adaptation iteration "<<iter<<" ---\n"<<endl;
      }
      int num = runOneIter();
      if ( verbosity >= 2 ) cout<<"\n";

      if ( num == 0 ) break;

      if ( iter == maxIterationsNumber  && num != 0 ) {
        MAdStatisticsSgl::instance().addInfiniteLoops(1);
        if ( verbosity >= 1 ) {
          cout << "Warning: reached maximum number of iterations\n";
        }
      }

      maxOp = std::max(maxOp,num);
      minOp = std::min(minOp,num);
    }

    // --- Geometry tracking ---
    if (geoTracker) {
      geoTracker->run();
      geoTracker->setHomogDirichletBC();
    }

    // --- Load balancing ---

#ifdef PARALLEL
    if( load_balance_algorithm != METIS_ALGORITHM ) {   
#ifdef _HAVE_PARMETIS_ 
      BalanceMetis2( mesh, *dataEchangerPtr );
#endif
    }
#endif
    cout << "\n +++ Ending a MDB mesh adaptation procedure +++\n";
    if ( verbosity >= 2 ) cout<<"\n";
  }

#ifdef PARALLEL
  // -------------------------------------------------------------------
  void MeshAdapter::setLoadBalancingAlgorithm( loadBalanceAlgorithm lbAlgo )
  {
    load_balance_algorithm = lbAlgo;
  }
  // -------------------------------------------------------------------
  void MeshAdapter::setDataExchanger( MDB_DataExchanger* dataExch_ptr )
  {
    dataEchangerPtr= dataExch_ptr;
  }
#endif

  // -------------------------------------------------------------------
  int MeshAdapter::partlyMoveObjects(double t, double dt, double* part)
  {
    return objects->partlyMove(*vMoveOp,t,dt,part);
  }

  // -------------------------------------------------------------------
  int MeshAdapter::moveObjectsAndReposition(double t, double dt, double chi)
  {
    return objects->moveAndReposition(mesh,t,dt,chi);
  }

  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
  void MeshAdapter::nodalDataDiagnostics(std::ostream& out) const
  {
    NodalDataManagerSgl::instance().diagnostics(out);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::openJournal() const
  {
    HistorySgl::instance().openJournal();
  }

  // -------------------------------------------------------------------
  void MeshAdapter::setReferenceJournal(std::string& name) const
  {
    HistorySgl::instance().loadJournal(name);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::flushJournal(std::ostream& out) const
  {
    HistorySgl::instance().flushJournal(out);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::enableSliverReports()
  {
    if ( !sliverROp ) {
      std::cerr<<"Error: no sliver region handler\n";
      exit(0);
    }

    sliverROp->enableReport(outPrefix);

    if ( !sliverFOp ) {
      std::cerr<<"Error: no sliver face handler\n";
      exit(0);
    }

    sliverFOp->enableReport(outPrefix);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::testSliverOperators(bool test)
  {
    sliverROp->setTestAllOperators(test);
  }

  // -------------------------------------------------------------------
  bool MeshAdapter::checkTheMesh(int verbosity, std::ostream& out, 
                                 MeshStatus * status) const
  {
    return checkMesh(mesh,CHECK_ALL,verbosity,out,status);
  }

  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
  // -------------------------------------------------------------------

  void MeshAdapter::setOutputPrefix(string prefix)
  {
    outPrefix = prefix;
  }

  // -------------------------------------------------------------------
  void MeshAdapter::writePos(string fn, MAdOutputData type) const
  {
    string fullName = outPrefix + fn;
    MAdGmshOutput(mesh, (const pSField) SFManager->getSizeField(), 
                  fullName.c_str(), type);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::writeMsh(string fn) const
  {
    string fullName = outPrefix + fn;
    M_writeMsh(mesh,fullName.c_str(),1,NULL);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::getStatistics(double * meanQuality, 
                                  double * worstQuality) const
  {
    MeshQualityManagerSgl::instance().evaluateStatistics();
    *meanQuality  = MeshQualityManagerSgl::instance().getMeanShape();
    *worstQuality = MeshQualityManagerSgl::instance().getWorstShape();
  }

  // -------------------------------------------------------------------
  void MeshAdapter::printStatistics(std::ostream& out) const
  {
    MeshQualityManagerSgl::instance().evaluateStatistics();
    MeshQualityManagerSgl::instance().printStatistics(out);
    out << "\n\n";
    MAdStatisticsSgl::instance().print(out);
  }

  // -------------------------------------------------------------------
  void MeshAdapter::printSliverRegionStatistics(std::ostream& out) const 
  {
    if ( sliverROp ) sliverROp->printStats(out);
    //   if ( sliverROp ) sliverROp->printOperatorsTest(out);
  }
  // -------------------------------------------------------------------
  void MeshAdapter::abort() const
  {
    cerr << "\n"
         << " ****************************\n"
         << "  MeshAdapter is aborting... \n"
         << " ****************************\n\n";    

    string stName = outPrefix + "aborted_statistics";
    std::ofstream stOut(stName.c_str());
    printStatistics(stOut);

    string slName = outPrefix + "aborted_slivers";
    std::ofstream slOut(slName.c_str());
    printSliverRegionStatistics(slOut);

    string jName = outPrefix + "aborted_journal";
    std::ofstream jOut(jName.c_str());
    flushJournal(jOut);

    writePos("aborted.pos",OD_MEANRATIO);
    writeMsh("aborted.msh");

    cerr << "Abort!\n";

    flush(cout);
    flush(cerr);

    exit(EXIT_FAILURE);
  }

  // -------------------------------------------------------------------

} // End of namespace MAd

