// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: Gaetan Compere, Jean-Francois Remacle
// -------------------------------------------------------------------

#include "MeanRatioEvaluator.h"

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
using std::string;

namespace MAd {

  // -------------------------------------------------------------------
  meanRatioEvaluator::meanRatioEvaluator(pSField f): 
    elementEvaluatorBase(f)
  {}

  // -------------------------------------------------------------------
  int meanRatioEvaluator::F_shape(pFace face, double normal[3],
                                  double * shape) const
  {
    double fxyz[3][3];
    pMSize pS[3];

    void *iter=0;
    int i = 0;
    pPList fvts=F_vertices(face,1);
    while( pVertex vt=(pVertex)PList_next(fvts,&iter) ) {
      pS[i]=sizeField->getSize(vt);
      V_coord(vt,fxyz[i++]);  
    }
    PList_delete(fvts);

    int res = XYZ_F_shape(fxyz,pS,normal,shape);  

    for (int idel=0; idel<3; idel++) if ( pS[idel] ) delete pS[idel];

    return res;
  }

  // -------------------------------------------------------------------
  int meanRatioEvaluator::XYZ_F_shape(double xyz[3][3], pMSize pS[3], 
                                      double nor[3], double * shape) const
  {
    if( sizeField->getType() == NULLSFIELD ) {
      if( ! XYZ_F_shape(xyz,pS[0],nor,shape) )
        { *shape=0; return 0; }
      return 1;
    }

    double shp[3];
    for(int i=0; i<3; i++) {
      if( ! XYZ_F_shape(xyz, pS[i], nor, &shp[i]) )
        { *shape = 0.; return 0; }
    }
    *shape = std::min(std::min(shp[0],shp[1]),shp[2]);

    return 1;
  }

  // -------------------------------------------------------------------
  int meanRatioEvaluator::XYZ_F_shape(double xyz[3][3], pMSize pS, 
                                      double nor[3], double * shape) const
  {
    double aSq,sumSq;
  
    aSq = sizeField->SF_XYZ_areaSq(xyz,pS,nor);

    if( aSq <= 0. ) { *shape = 0.; return 0; } 
  
    sumSq  = sizeField->SF_XYZ_lengthSq(xyz[1],xyz[0],pS);
    sumSq += sizeField->SF_XYZ_lengthSq(xyz[2],xyz[0],pS);
    sumSq += sizeField->SF_XYZ_lengthSq(xyz[1],xyz[2],pS);
    
    *shape = 48.*aSq/(sumSq*sumSq);

    if(*shape < acptValue) return 0;
    else if (*shape > (1.+MAdTOL) ) *shape = 1.;

    return 1;
  }

  // -------------------------------------------------------------------
  int meanRatioEvaluator::R_shape(pRegion rgn, double * shape) const
  {
    double rxyz[4][3];
    pMSize pS[4];
  
    void *iter=0;
    int i=0;
    pPList verts=R_vertices(rgn);
    while( pVertex vt=(pVertex)PList_next(verts,&iter) ) {
      pS[i]=sizeField->getSize(vt);
      V_coord(vt,rxyz[i++]);
    }  
    PList_delete(verts);

    int res = XYZ_R_shape(rxyz,pS,shape);

    for (int idel=0; idel<4; idel++) if ( pS[idel] ) delete pS[idel];  

    return res;
  }

  // -------------------------------------------------------------------
  int meanRatioEvaluator::XYZ_R_shape(double xyz[4][3],pMSize pS[4], 
                                      double * shape) const
  {
    if( sizeField->getType() == NULLSFIELD ) {
      if( ! XYZ_R_shape(xyz,pS[0],shape) )
        { *shape=0.; return 0; }
      return 1;
    }

    double hmin = MAdBIG;
    pMSize pSZ;

    // use worst
    for(int i=0; i<4; i++) {
      if( pS[i] ) {
        if( pS[i]->size(0) <  hmin )
          { hmin=pS[i]->size(0); pSZ=pS[i]; }
      }
    }
    if( hmin != MAdBIG ) {
      if( ! XYZ_R_shape(xyz,pSZ,shape) )
        return 0; 
    }
    else
      throw;
  
    return 1;
  }

  // -------------------------------------------------------------------
  int meanRatioEvaluator::XYZ_R_shape(double xyz[4][3], pMSize pS, 
                                      double * shape) const
  {
    double vol,sumSq;
  
    vol = sizeField->SF_XYZ_volume(xyz,pS);

    if ( vol < 0. ) { *shape = 0.; return 0; }
  
    sumSq  = sizeField->SF_XYZ_lengthSq(xyz[1],xyz[0],pS);
    sumSq += sizeField->SF_XYZ_lengthSq(xyz[2],xyz[0],pS);
    sumSq += sizeField->SF_XYZ_lengthSq(xyz[3],xyz[0],pS);
    sumSq += sizeField->SF_XYZ_lengthSq(xyz[3],xyz[1],pS);
    sumSq += sizeField->SF_XYZ_lengthSq(xyz[3],xyz[2],pS);
    sumSq += sizeField->SF_XYZ_lengthSq(xyz[1],xyz[2],pS);

    *shape = 15552.*vol*vol/(sumSq*sumSq*sumSq);

    if(*shape < acptValue) return 0;
    else if (*shape > (1.+MAdTOL) ) *shape = 1.;

    return 1;
  }

  // -------------------------------------------------------------------

}
