// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: Gaetan Compere, Jean-Francois Remacle
// -------------------------------------------------------------------

#include "GeoMatcher.h"
#include "CallBackManager.h"
#include "MathUtils.h"
#include "ModelInterface.h"

namespace MAd {

  // -------------------------------------------------------------------
  void GeoMatcherCBFunction (pPList before, pPList after, void * data,
                             operationType type , pEntity ppp) 
  {
#ifdef _HAVE_GMSH_
    geoMatcher * gm = (geoMatcher *)(data);

    switch (type) {
    case MAd_ESPLIT: { 
      // if new point is on a boundary, you need to find its 
      // projection on the boundary and impose its new 
      // location as a Dirichlet BC in the elastic model

      pGEntity pGE = V_whatIn( (pVertex)ppp );
      int dimBnd = GEN_type(pGE);
      if ( dimBnd < gm->meshDim() ) { // new point is on boundary

        double xyz[3]; // current position of the new point
        V_coord( (pVertex)ppp, xyz );
        printVec(xyz,"original location");

        double u,v;
        if ( !V_params((pVertex)ppp,&u,&v) ) {
          printf("Error in GeoMatcherCBFunction: not a parametric point\n");
          throw;
        }

        double tgt[3]; // the target location
        switch (dimBnd) {
        case 2: { // on a surface
          GF_xyz((pGFace)pGE, u, v, tgt);
          break;
        }
        case 1: { // on a line
          GE_xyz((pGEdge)pGE, u, tgt);
          break;
        }
        default: {
          printf("Error in GeoMatcherCBFunction: unexpected geometric entity dimension for a new vertex: %d\n",
                 dimBnd);
          throw;
        }
        }
        printVec(tgt,"target location");

        // compute the difference with the current position
        double dx[3];
        diffVec(tgt,xyz,dx);
      
        // impose the motion as a Dirichlet BC in the elastic model
        smallVector dxVec(3);
        for (int i=0; i<3; i++) dxVec(i) = dx[i];
        gm->addDirichlet( (pVertex)ppp, dxVec );
      }
      break;
    } 
    case MAd_ECOLLAPSE: {
      gm->delDirichlet( (pVertex)ppp );
      break;
    }
    case MAd_FSWAP:
    case MAd_ESWAP: {
      break;
    }
    case MAd_RREMOVE: {
      void * temp = NULL;
      while ( pEntity pE = PList_next(before,&temp) ) {
        if ( EN_type(pE) == 0 ) {
          gm->delDirichlet( (pVertex)pE );
        }
      }
      break;
    }
    default: {
      printf("Error in GeoMatcherCBFunction: not implemented for mesh modification %d",
             type);  
      throw;
    }
    }
#endif
  };

  // -------------------------------------------------------------------
  geoMatcher::geoMatcher(pMesh _mesh):
    MAdElasticityOp(_mesh)
  {
    CallBackManagerSgl::instance().registerCallBack(GeoMatcherCBFunction,this);
  }

  // -------------------------------------------------------------------
  geoMatcher::geoMatcher(const geoMatcher& gm):
    MAdElasticityOp(gm)
  {
    CallBackManagerSgl::instance().registerCallBack(GeoMatcherCBFunction,this);
  }

  // -------------------------------------------------------------------
  geoMatcher::~geoMatcher()
  {
    CallBackManagerSgl::instance().unregisterCallBack(GeoMatcherCBFunction,this);
  }

  // -------------------------------------------------------------------

}
