// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: Gaetan Compere, Jean-Francois Remacle
// -------------------------------------------------------------------

#include "MAdElasticityOp.h"
#include "CPUTimeManager.h"
#include "OperatorTools.h"
#include "MAdLinearSystem.h"
#include "MathUtils.h"

#include <iostream>
using std::cout;
using std::cerr;
using std::endl;
using std::make_pair;
using std::map;
using std::set;

namespace MAd {

  // -------------------------------------------------------------------
  MAdElasticityOp::MAdElasticityOp(pMesh m):
    mesh(m),E(1.),nu(0.45),chi(1.0)
  {
    dim = M_dim(mesh);
    if (dim!=3) {
      cerr << "Error: elasticity model only implemented for dimension 3, current: " << dim << endl;
      throw;
    }
  }

  // -------------------------------------------------------------------
  MAdElasticityOp::MAdElasticityOp(const MAdElasticityOp& eop):
    mesh(eop.mesh), dim(eop.dim), 
    E(eop.E), nu(eop.nu), chi(eop.chi)
  {
    localIds = eop.localIds;
    dirichlet = eop.dirichlet;
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::setMaterials(double _E, double _nu)
  {
    E = _E;  nu = _nu;
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::setStiffnessAlterationCoef(double _chi)
  {
    chi = _chi;
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::setHomogDirichletBC()
  {
    dirichlet.clear();
    smallVector zero(3);  zero.set_all(0.);

    set<pVertex> verts;
    collectBCVertices(&verts);

    set<pVertex>::const_iterator vIter = verts.begin();
    set<pVertex>::const_iterator vLast = verts.end();
    for (; vIter != vLast; vIter++) {
      addDirichlet(*vIter,zero);
    }
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::collectBCVertices(set<pVertex> * bcVerts)
  {
    FIter fit = M_faceIter(mesh);
    while ( pFace pf = FIter_next(fit) ) {
      if ( EN_whatInType((pEntity)pf) == 2 ) {
        pPList fV = F_vertices(pf,1);
        void* tmp=0;
        while ( pVertex pv = (pVertex)PList_next(fV,&tmp) ) {
          (*bcVerts).insert(pv);
        }
        PList_delete(fV);
      }
    }
    FIter_delete(fit);
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::addDirichlet(pVertex pv, smallVector& dxyz)
  {
    map<pVertex,smallVector>::iterator dIter = dirichlet.find(pv);
    if ( dIter != dirichlet.end() ) {
      smallVector dxyz1 = (*dIter).second;
      dxyz1.add(dxyz);
      dirichlet.erase(dIter);
      dirichlet.insert(make_pair(pv,dxyz1));
    }
    else {
      dirichlet.insert(make_pair(pv,dxyz));
    }
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::delDirichlet(pVertex pv)
  {
    map<pVertex,smallVector>::iterator dIter = dirichlet.find(pv);
    if ( dIter != dirichlet.end() ) {
      dirichlet.erase(dIter);
    }
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::generateVertexIds()
  {
    localIds.clear();

    int id = 0;

    VIter vit = M_vertexIter(mesh);
    while ( pVertex pv = VIter_next(vit) ) {
      localIds[pv] = id;
      id++;
    }
    VIter_delete(vit);
  }

  // -------------------------------------------------------------------
  int MAdElasticityOp::run()
  {
    CPUTimeManager& tm = CPUTimeManagerSgl::instance();

    MeshQualityManagerSgl::instance().evaluateSizes(); // required for detJ0

    // --- produce the mapping pVertex -> localId ---
    generateVertexIds();

    // --- create the linear system ---
    double t_sys0 = tm.getTime();
    int nbVert = M_numVertices(mesh);
    int sysSize = dim*nbVert;
    MAdLinearSystemDef lsys = MAdLinearSystemDef();
    lsys.allocate(sysSize);

    VIter vit2 = M_vertexIter(mesh);
    while ( pVertex pv2 = VIter_next(vit2) ) {
      int locId = localIds[pv2];
      int nbNeighbors = V_numEdges(pv2);
      for (int iDir=0; iDir<3; iDir++) {
        int row = nbVert*iDir + locId;
        lsys.set_nnz(row,(nbNeighbors+1)*3);// A node is coupled with all its neighbors, and don't forget himself
      }
    }
    VIter_delete(vit2);
    lsys.allocate_matrix();

    lsys.zeroMatrix();
    lsys.zeroRightHandSide();

    lsys.setSolver(CG);
    lsys.setFillIn(4);
    lsys.setEps(1.e-12);

    lsys.setPrec(1.e-12); // for gmm only
    lsys.setNoisy(1);    // for gmm only

    double dt_sys = tm.getTime() - t_sys0;
    cout << "Created the system in "<<dt_sys<<" seconds\n";

    // --- assemble the matrix ---
    int test=0;
    double t_ass0 = tm.getTime();
    RIter rit = M_regionIter(mesh);
    while ( pRegion pr = RIter_next(rit) ) {

      int rNodes = R_numVertices(pr);
      int locSyze = rNodes*dim;

      smallMatrix locK(locSyze,locSyze);
      element3DMatrix(pr, locK);

      pPList vertsI = R_vertices(pr);
      pPList vertsJ = R_vertices(pr);
      void* tmpI = 0;
      int i = 0;
      while ( pVertex pvI = (pVertex)PList_next(vertsI,&tmpI) ) {

        int locIdI = localIds[pvI];

        void* tmpJ = 0;
        int j = 0;
        while ( pVertex pvJ = (pVertex)PList_next(vertsJ,&tmpJ) ) {

          int locIdJ = localIds[pvJ];
          
          for (int k=0; k<dim; k++) {
            for (int l=0; l<dim; l++) {
              lsys.addToMatrix(nbVert*k + locIdI, nbVert*l+locIdJ, locK(k*rNodes+i,l*rNodes+j));
            }
          }
          j++;
        }
        i++;
      }
      PList_delete(vertsI);
      PList_delete(vertsJ);
      test++;
    }
    RIter_delete(rit);
    double dt_ass = tm.getTime() - t_ass0;
    cout << "Assembled the matrix in "<<dt_ass<<" seconds\n";

    // --- apply bc (ugly but it works hum) ---
    double t_bc0 = tm.getTime();
    map<pVertex,smallVector>::iterator dIter = dirichlet.begin();
    map<pVertex,smallVector>::iterator dLast = dirichlet.end();
    for (; dIter != dLast; dIter++) {

      pVertex pv       = (*dIter).first;
      smallVector dxyz = (*dIter).second;

      // get the vertex id
      map<pVertex,int>::const_iterator itId = localIds.find(pv);
      int id = (*itId).second;
    
      for (int iDir=0; iDir<3; iDir++) {
        int row = nbVert*iDir + id;
        const double BIG = /*lsys.getFromMatrix(row,row) *  */1.e12;
        lsys.addToMatrix(row,row,BIG);
        lsys.addToRightHandSide(row,BIG * dxyz(iDir));
      }
    
    }
    double dt_bc = tm.getTime() - t_bc0;
    cout << "Applied boundary conditions in "<<dt_bc<<" seconds\n";

    // --- reorder matrix ---
    lsys.reorder();

    // --- solve the system ---
    double t_slv0 = tm.getTime();
    lsys.systemSolve();
    double dt_slv = tm.getTime() - t_slv0;
    cout << "Solved the elastic system in "<<dt_slv<<" seconds\n";

    // --- export the result to the mesh ---
    VIter vit = M_vertexIter(mesh);
    while ( pVertex pv = VIter_next(vit) ) {

      // get the vertex id
      map<pVertex,int>::const_iterator itId = localIds.find(pv);
      int id = (*itId).second;

      // get the dx
      double dxyz[3];
      for (int iDir=0; iDir<3; iDir++) {
        int row = nbVert*iDir + id;
        dxyz[iDir] = lsys.getFromSolution(row);
      }

      // compute the new position
      double xyz0[3];
      V_coord(pv,xyz0);
      double xyz[3];
      for (int iDir=0; iDir<3; iDir++) xyz[iDir] = xyz0[iDir] + dxyz[iDir];

      // move the vertex
      V_setPosition(pv,xyz);
    }
    VIter_delete(vit);
  
    localIds.clear();
    dirichlet.clear();

    return 1;
  }

  // -------------------------------------------------------------------
  void MAdElasticityOp::element3DMatrix(pRegion pr, smallMatrix& m) const
  {
    int nbNodes = 4;
    double xyz[4][3];
    R_coordP1(pr,xyz);

    double FACT = E / (1 + nu);
    double C11 = FACT * (1 - nu) / (1 - 2 * nu);
    double C12 = FACT * nu / (1 - 2 * nu);
    double C44 = (C11 - C12) / 2;
    const double C[6][6] =
      { {C11, C12, C12,    0,   0,   0}, 
        {C12, C11, C12,    0,   0,   0}, 
        {C12, C12, C11,    0,   0,   0}, 
        {  0,   0,   0,  C44,   0,   0},
        {  0,   0,   0,    0, C44,   0}, 
        {  0,   0,   0,    0,   0, C44} };

    smallMatrix H   (6,6);
    for (int i=0;i<6;i++)
      for (int j=0;j<6;j++)
        H(i,j) = C[i][j];

    smallMatrix B   (6,3*nbNodes);
    smallMatrix BTH (3*nbNodes,6);
    smallMatrix BT  (3*nbNodes,6);

    const double gradShpFct[4][3] =
      { {-1., -1., -1.},
        { 1.,  0.,  0.},
        { 0.,  1.,  0.},
        { 0.,  0.,  1.} };

    double jac [3][3];
    jac[0][0] = jac[0][1] = jac[0][2] = 0.;
    jac[1][0] = jac[1][1] = jac[1][2] = 0.;
    jac[2][0] = jac[2][1] = jac[2][2] = 0.;
    for(int i = 0; i < nbNodes; i++) {
      jac[0][0] += xyz[i][0] * gradShpFct[i][0]; jac[0][1] += xyz[i][1] * gradShpFct[i][0]; jac[0][2] += xyz[i][2] * gradShpFct[i][0];
      jac[1][0] += xyz[i][0] * gradShpFct[i][1]; jac[1][1] += xyz[i][1] * gradShpFct[i][1]; jac[1][2] += xyz[i][2] * gradShpFct[i][1];
      jac[2][0] += xyz[i][0] * gradShpFct[i][2]; jac[2][1] += xyz[i][1] * gradShpFct[i][2]; jac[2][2] += xyz[i][2] * gradShpFct[i][2];
    }
    const double detJ = fabs(jac[0][0] * jac[1][1] * jac[2][2] + jac[0][2] * jac[1][0] * jac[2][1] +
                             jac[0][1] * jac[1][2] * jac[2][0] - jac[0][2] * jac[1][1] * jac[2][0] -
                             jac[0][0] * jac[1][2] * jac[2][1] - jac[0][1] * jac[1][0] * jac[2][2]);

    double invjac [3][3];
    inverseMat (jac, invjac) ;

    double Grads[4][3];

    B.set_all(0.0);
    BT.set_all(0.0);

    for (int j=0;j<nbNodes;j++){

      Grads[j][0] = invjac[0][0] * gradShpFct[j][0] + invjac[0][1] * gradShpFct[j][1] + invjac[0][2] * gradShpFct[j][2];
      Grads[j][1] = invjac[1][0] * gradShpFct[j][0] + invjac[1][1] * gradShpFct[j][1] + invjac[1][2] * gradShpFct[j][2];
      Grads[j][2] = invjac[2][0] * gradShpFct[j][0] + invjac[2][1] * gradShpFct[j][1] + invjac[2][2] * gradShpFct[j][2];
    
      BT(j,0) = B(0,j) = Grads[j][0];
      BT(j,3) = B(3,j) = Grads[j][1];
      BT(j,4) = B(4,j) = Grads[j][2];
    
      BT(j+nbNodes,1) = B(1,j+nbNodes) = Grads[j][1];
      BT(j+nbNodes,3) = B(3,j+nbNodes) = Grads[j][0];
      BT(j+nbNodes,5) = B(5,j+nbNodes) = Grads[j][2];
    
      BT(j+2*nbNodes,2) = B(2,j+2*nbNodes) = Grads[j][2];
      BT(j+2*nbNodes,4) = B(4,j+2*nbNodes) = Grads[j][0];
      BT(j+2*nbNodes,5) = B(5,j+2*nbNodes) = Grads[j][1];
    }
  
    BTH.set_all(0.0);
    BTH.blas_dgemm (BT,H); 
  
    m.set_all(0.0);
    m.blas_dgemm (BTH,B,detJ,1.0);

    // --- Selective element stiffening ---
    // reference:
    // Stein, Tezduyar, Benney "Mesh Moving Techniques for FSI with Large Displacement", ASME, 2003.

    double detJ0 = MeshQualityManagerSgl::instance().getMaxSize();
    if ( detJ0 <= 0. ) detJ0 = 1.;

    double scaleFactor = pow( detJ0 / detJ, chi);
    if ( scaleFactor > 1.e12 || scaleFactor < 1.e-12 ) {
      cout <<"Warning in MAdElasticity: scale factor for element stiffness is "<<scaleFactor<<endl;
    }

    m.scale ( scaleFactor );
  }

  // -------------------------------------------------------------------

}
