// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: Gaetan Compere, Jean-Francois Remacle
// -------------------------------------------------------------------

#include "AnisoMeshSize.h"
#include "MathUtils.h"

#include <iostream>
using std::string;

namespace MAd {

  // -------------------------------------------------------------------
  AnisoMeshSize::AnisoMeshSize():
    MeshSizeBase()
  {}

  // -------------------------------------------------------------------
  AnisoMeshSize::AnisoMeshSize(double _h[3], 
                               double eDiag[3], 
                               double eOffDiag[3]):
    MeshSizeBase()
  {
    double _e[3][3];
    _e[0][0] = eDiag[0];
    _e[1][1] = eDiag[1];
    _e[2][2] = eDiag[2];

    _e[1][0] = eOffDiag[0];
    _e[2][0] = eOffDiag[1];
    _e[2][1] = eOffDiag[2];

    _e[0][1] = eOffDiag[0];
    _e[0][2] = eOffDiag[1];
    _e[1][2] = eOffDiag[2];

    processMetric(_e,_h);  
  }

  // -------------------------------------------------------------------
  AnisoMeshSize::AnisoMeshSize(double dirs[3][3], double _h[3]):
    MeshSizeBase()
  {
    processMetric(dirs,_h);
  }

  // -------------------------------------------------------------------
  AnisoMeshSize::AnisoMeshSize(double _h):
    MeshSizeBase()
  {
    e[0][0] = 1.0;
    e[0][1] = 0.0;
    e[0][2] = 0.0;
    e[1][0] = 0.0;
    e[1][1] = 1.0;
    e[1][2] = 0.0;
    e[2][0] = 0.0;
    e[2][1] = 0.0;
    e[2][2] = 1.0;
    h[0] = _h;
    h[1] = _h;
    h[2] = _h;
  }

  // -------------------------------------------------------------------
  AnisoMeshSize::AnisoMeshSize(AnisoMeshSize *pm):
    MeshSizeBase()
  {
    e[0][0] = pm->e[0][0];
    e[0][1] = pm->e[0][1];
    e[0][2] = pm->e[0][2];
    e[1][0] = pm->e[1][0];
    e[1][1] = pm->e[1][1];
    e[1][2] = pm->e[1][2];
    e[2][0] = pm->e[2][0];
    e[2][1] = pm->e[2][1];
    e[2][2] = pm->e[2][2];
    h[0] = pm->h[0];
    h[1] = pm->h[1];
    h[2] = pm->h[2];
  }

  // -------------------------------------------------------------------
  MeshSizeType AnisoMeshSize::getType() const 
  {
    return ANISOTROPIC;
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::processMetric(const double dirs[3][3], const double hs[3])
  {
    printf("AnisoMeshSize::processMetric not implemented\n");
    throw;
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::intersect(const pMSize pMS0, const pMSize pMS1)
  {
    printf("AnisoMeshSize::intersect not implemented\n");
    throw;
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::interpolate(const pMSize pMS0, const pMSize pMS1, 
                                  double param)
  {
    printf("AnisoMeshSize::interpolate not implemented\n");
    throw;
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::direction(int i, double dir[3]) const 
  {
    for (int c=0; c<3; c++) dir[c] = e[i][c];
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::scale(int dir, double factor)
  {
    h[dir] *= factor;
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::scale(double factor)
  {
    h[0] *= factor;
    h[1] *= factor;
    h[2] *= factor;
  }

  // -------------------------------------------------------------------
  // get the square of desired edge length along 'vec'
  double AnisoMeshSize::lengthSqInDir(double vec[3]) const
  {
    double vecTr[3];
    toMetricSpace(vec,vecTr);
    return dotProd(vec,vec)/dotProd(vecTr,vecTr);
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::toMetricSpace(const double vec[3], double vecTr[3]) const
  {
    vecTr[0] = dotProd(vec,e[0])/h[0];
    vecTr[1] = dotProd(vec,e[1])/h[1];
    vecTr[2] = dotProd(vec,e[2])/h[2];
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::computeMetric(double metr[3][3]) const
  {
    // M = e^t * diag * e = e * diag * e^t
    // where the elements of diag are lambda_i = 1 / h_i^2

    double tmp[3][3];
    double lam[3];
    for (int i=0; i<3; i++)   lam[i] = 1. / ( h[i] * h[i] );
    tmp[0][0] = lam[0] * e[0][0];
    tmp[0][1] = lam[1] * e[1][0];
    tmp[0][2] = lam[2] * e[2][0];
    tmp[1][0] = lam[0] * e[0][1];
    tmp[1][1] = lam[1] * e[1][1];
    tmp[1][2] = lam[2] * e[2][1];
    tmp[2][0] = lam[0] * e[0][2];
    tmp[2][1] = lam[1] * e[1][2];
    tmp[2][2] = lam[2] * e[2][2];

    matMat(tmp,e,metr);
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::print(string name) const
  {
    std::cout<<"Printing AnisoMeshSize \'"<<name<<"\' ("<<this<<")\n";
    std::cout<<"  h:\t"<<h[0]<<"\t"<<h[1]<<"\t"<<h[2]<<"\n";
    std::cout<<"  e:\n";
    for (int i = 0; i<3; i++) {
      std::cout << "    "<<e[0][i]<<"\t"<<e[1][i]<<"\t"<<e[2][i]<<"\n";
    }  
  }

  // -------------------------------------------------------------------

}
