// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: Gaetan Compere, Jean-Francois Remacle
// -------------------------------------------------------------------

#include "LocalSizeField.h"
#include "IsoMeshSize.h"
#include "AnisoMeshSize.h"
#include "MathUtils.h"

#include <iostream>
using std::cerr;
using std::cout;
using std::endl;
using std::set;

namespace MAd {

  // -------------------------------------------------------------------
  LocalSizeField::LocalSizeField(pMesh m):
    SizeFieldBase(), mesh(m), 
    radius(-1.), sizeIn(-1.), scaleAtRadius(-1.)
  {
    kdSearch = new SearchTool();
  }
  
  // -------------------------------------------------------------------
  LocalSizeField::LocalSizeField(pMesh m, double rad, double _sizeIn, double scale):
    SizeFieldBase(), mesh(m)
  { 
    setSizes(rad, _sizeIn, scale);
    kdSearch = new SearchTool();
  }
  
  // -------------------------------------------------------------------
  LocalSizeField::LocalSizeField(const LocalSizeField& _lsf)
  { 
    throw;
  }
  
  // -------------------------------------------------------------------
  LocalSizeField::~LocalSizeField()  
  {
    if (kdSearch) delete kdSearch;
  }

  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
  void LocalSizeField::setSizes(double rad, double _sizeIn, double _scale)
  {
    radius = rad; 
    sizeIn = _sizeIn; 
    scaleAtRadius = _scale;
  }

  // -------------------------------------------------------------------
  void LocalSizeField::addGeometricEntity(int type, int tag)
  {
    geomEntities.insert( std::make_pair(type,tag) );
  }

  // -------------------------------------------------------------------
  // -------------------------------------------------------------------
  pMSize LocalSizeField::getSize(const pVertex pv) const
  {
    double xyz[3];
    V_coord(pv,xyz);
    double dist = getDistance(xyz);
    return getSize(dist);
  }

  // ----------------------------------------------------------------------
  pMSize LocalSizeField::getSize(double dist) const
  {
    pMSize theSize;

    if (dist > radius) {
      theSize = new IsoMeshSize (MAdBIG);
    }
    else {
      double currentScale = 1. + ( dist / radius ) * ( scaleAtRadius - 1. );
      theSize = new IsoMeshSize(sizeIn*currentScale);
    }
    return theSize;
  }

  // -------------------------------------------------------------------
  pMSize LocalSizeField::getSizeOnEntity(const pEntity, 
                                         const double xyz[3]) const
  {
    double dist = getDistance(xyz);
    return getSize(dist);
  }

  // -------------------------------------------------------------------
  // Length squared computation
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  double LocalSizeField::SF_VV_lengthSq(const pVertex pV0, const pVertex pV1) const
  {
    double xyz[2][3];
    V_coord(pV0,xyz[0]);
    V_coord(pV1,xyz[1]);

    pMSize pS[2];
    pS[0] = getSize(pV0);
    pS[1] = getSize(pV1);

    double lSq = SF_XYZ_lengthSq(xyz[0],xyz[1],pS[0],pS[1]);
  
    if ( pS[0] ) delete pS[0];
    if ( pS[1] ) delete pS[1];

    return lSq;
  }

  // -------------------------------------------------------------------
  double LocalSizeField::SF_XYZ_lengthSq(const double xyz0[3], 
                                         const double xyz1[3], 
                                         const pMSize pS0, 
                                         const pMSize pS1) const
  {
    if( pS0 ) {

      double e[3];
      diffVec(xyz0,xyz1,e);
    
      double eTr[3];
      pS0->toMetricSpace(e,eTr);
      double lenSq0 = dotProd(eTr,eTr);

      if ( pS1 )
        {
          pS1->toMetricSpace(e,eTr);
          double lenSq1 = dotProd(eTr,eTr);
          return sqrt(lenSq0*lenSq1);
        }
      else return lenSq0;
    }
    else {
      printf("Error in LocalSizeField::lengthSq: no size defined\n");
      throw;
    }
    return 0.;
  }

  // -------------------------------------------------------------------
  // Area squared computation
  // -------------------------------------------------------------------

  double LocalSizeField::SF_F_areaSq(const pFace face) const
  {
    double area = 0.;

    double xyz[3][3];
    F_coordP1(face,xyz);
  
    void * temp = 0;
    pPList fVerts = F_vertices(face,1);
    while( pVertex pV = (pVertex)PList_next(fVerts,&temp) )
      {
        pMSize pS = getSize(pV);
        area += SF_XYZ_areaSq(xyz,pS,0);
        if (pS) delete pS;
      }
    PList_delete(fVerts);
  
    area /= F_numVertices(face);

    return area;
  }

  // -------------------------------------------------------------------
  double LocalSizeField::SF_XYZ_areaSq(const double fxyz[3][3], 
                                       const pMSize pS, 
                                       const double norDir[3]) const
  {
    if( !pS ) {
      printf("Error in LocalSizeField::areaSq: no size given\n");
      throw;
    }

    // get the two first edges
    double e01[3],e02[3];
    diffVec(fxyz[1],fxyz[0],e01);
    diffVec(fxyz[2],fxyz[0],e02);

    double e01Tr[3],e02Tr[3];
    pS->toMetricSpace(e01,e01Tr);
    pS->toMetricSpace(e02,e02Tr);

    double nor[3];
    crossProd(e01Tr,e02Tr,nor);

    if( norDir && dotProd(norDir,nor) < MAdTOL ) return 0.;

    double area = 0.25 * dotProd(nor,nor);
    if( area < MAdTOL ) return 0.;

    return area;
  }

  // -------------------------------------------------------------------
  // Volume computation
  // -------------------------------------------------------------------

  double LocalSizeField::SF_R_volume(const pRegion region) const
  {
    double vol = 0.;

    double xyz[4][3];
    R_coordP1(region,xyz);

    pPList rVerts = R_vertices(region);
    void * temp = 0;
    while( pVertex pV = (pVertex)PList_next(rVerts,&temp) )
      {
        pMSize pS = getSize(pV);
        vol += SF_XYZ_volume(xyz,pS);
        if (pS) delete pS;
      }
    PList_delete(rVerts);

    vol /= R_numVertices(region);

    return vol;
  }

  // -------------------------------------------------------------------
  double LocalSizeField::SF_XYZ_volume(const double xyz[4][3], const pMSize pS) const
  {
    if( !pS ) {
      printf("Error in LocalSizeField::volume: no size given\n");
      throw;
    }

    double physVol = R_XYZ_volume(xyz);
  
    return ( physVol / (pS->size(0)*pS->size(1)*pS->size(2)) );
  }

  // -------------------------------------------------------------------
  // Center of edge computation
  // -------------------------------------------------------------------

  // -------------------------------------------------------------------
  double LocalSizeField::SF_E_center(const pEdge edge, double center[3], 
                                     pMSize * cSize) const
  {
    return SF_VV_center(E_vertex(edge,0),E_vertex(edge,1),center,cSize);
  }

  // -------------------------------------------------------------------
  double LocalSizeField::SF_VV_center(const pVertex v0, const pVertex v1,
                                      double center[3], pMSize * cSize) const
  {
    double xyz[2][3];
    V_coord(v0,xyz[0]);
    V_coord(v1,xyz[1]);

    pMSize pS[2];
    pS[0] = getSize(v0);
    pS[1] = getSize(v1);

    double cParam = SF_XYZ_center(xyz,pS,center,cSize);

    if ( pS[0] ) delete pS[0];
    if ( pS[1] ) delete pS[1];

    return cParam;
  }

  // ----------------------------------------------------------------------
  // -------------------------------------------------------------------
  void LocalSizeField::scale(double fact)
  {
    printf("Not implemented (LocalSizeField::scale)\n");
    throw;  
  }

  // ----------------------------------------------------------------------
  // ----------------------------------------------------------------------
  void LocalSizeField::updateTree()
  {
    kdSearch->reset();

    set<pVertex> vertices = collectVertices();
    int nbVert = vertices.size();

    kdSearch->allocatePoints(nbVert);

    set<pVertex> :: iterator itV    = vertices.begin();
    set<pVertex> :: iterator itVEnd = vertices.end();
    int k = 0;
    for ( ; itV != itVEnd ; ++itV)
      {
        double coord[3]; V_coord(*itV,coord);
        kdSearch->addPoint(k,coord);
        k++;
      }

    kdSearch->allocateTree(nbVert);
  }

  // ----------------------------------------------------------------------
  set<pVertex> LocalSizeField::collectVertices() const
  {
    set<pVertex> vertices;

    set<std::pair<int,int> >::const_iterator it = geomEntities.begin();
    set<std::pair<int,int> >::const_iterator itEnd = geomEntities.end();
    for (; it != itEnd; it++)  addVerticesOnGEntity(it->first, it->second, &vertices);

    return vertices;
  }

  // ----------------------------------------------------------------------
  void LocalSizeField::addVerticesOnGEntity(int type, int tag, set<pVertex> * verts) const
  {
    switch (type) {
    case 0:
      {
        VIter vit = M_vertexIter(mesh);
        while (pVertex pv = VIter_next(vit))
          {
            pGEntity pg = EN_whatIn((pEntity)pv);
            int pgType = EN_whatInType( (pEntity) pv );
            if(pg)
              if (GEN_tag(pg) == tag && pgType == 0)
                {
                  verts->insert(pv);
                }
          }
        VIter_delete(vit);
        break;
      }
    case 1:
      {
        EIter eit = M_edgeIter(mesh);
        while (pEdge pe = EIter_next(eit))
          {
            pGEntity pg = EN_whatIn((pEntity)pe);
            int pgType = EN_whatInType( (pEntity) pe );
            if(pg)
              if (GEN_tag(pg) == tag && pgType == 1)
                {
                  verts->insert(E_vertex (pe,0));
                  verts->insert(E_vertex (pe,1));
                }
          }
        EIter_delete(eit);
      }
    case 2:
      {
        FIter fit = M_faceIter(mesh);
        while (pFace pf = FIter_next(fit))
          {
            pGEntity pg = EN_whatIn((pEntity)pf);
            int pgType = EN_whatInType( (pEntity) pf );
            if(pg)
              if (GEN_tag(pg) == tag && pgType == 2)
                {
                  verts->insert(F_vertex (pf,0));
                  verts->insert(F_vertex (pf,1));
                  verts->insert(F_vertex (pf,2));
                }
          }
        FIter_delete(fit);
      }
    case 3:
      {
        RIter rit = M_regionIter(mesh);
        while (pRegion pr = RIter_next(rit))
          {
            pGEntity pg = EN_whatIn((pEntity)pr);
            int pgType = EN_whatInType( (pEntity) pr );
            if(pg)
              if (GEN_tag(pg) == tag && pgType == 3)
                {
                  verts->insert(R_vertex (pr,0));
                  verts->insert(R_vertex (pr,1));
                  verts->insert(R_vertex (pr,2));
                  verts->insert(R_vertex (pr,3));
                }
          }
        RIter_delete(rit);
      }
    }
  }

  // ----------------------------------------------------------------------
  double LocalSizeField::getDistance(const double xyz[3]) const 
  {
    double distSq = kdSearch->computeDistanceSq(xyz);
    return sqrt(distSq);
  }

  // ----------------------------------------------------------------------

}
