// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: Gaetan Compere, Jean-Francois Remacle
// -------------------------------------------------------------------

// the template parameter file for this executable
#include "Parameters.h"
// the parameter file for the current test case
#include "MyParams.h"

#include "MSops.h"
#include "AdaptInterface.h"
#include "NullSField.h"

#include <iostream>
#include <sstream>
#include <math.h>
#include <sys/time.h>
#include <fstream>
#include <stdlib.h>
#include <vector>
#include <set>

#ifdef PARALLEL
 #include "mpi.h"
#endif

using std::cout;
using std::cerr;
using std::endl;
using std::stringstream;
using std::ofstream;
using std::vector;
using std::set;

using namespace MAd;

// ----------------------------------------------------------------------
double CPUTime() {
  
#ifdef PARALLEL
  return MPI_Wtime();
#else
  struct timeval tp;
  struct timezone tz;

  gettimeofday(&tp,&tz);

  return ((double) tp.tv_sec +
          (double) ((double) .000001 * (double) tp.tv_usec));
#endif
}

// ----------------------------------------------------------------------
void writeSolution(MeshAdapter* ma, pMesh mesh, int iter, string type) 
{
  stringstream ss;
  string iterStr;  ss << iter;  ss >> iterStr;
  
  if ( !strcmp(type.c_str(),"Pos") || !strcmp(type.c_str(),"MshAndPos") ) {
    string namePos = "result" + iterStr + ".pos";
    ma->writePos(namePos,OD_MEANRATIO);
  }
  
  if ( !strcmp(type.c_str(),"Msh") || !strcmp(type.c_str(),"MshAndPos") ) {
    string nameMsh = "result" + iterStr + ".msh";
    ma->writeMsh(nameMsh);
  }
}

// ----------------------------------------------------------------------
void setMobileObject(mobileObject* mob, pMesh mesh, ObjectDef def)
{
  // --- Get name ---
  string name = def.name;
  mob->setName(name);

  // --- Geometry ---
  int type = def.geomLevel;
  set<int>::const_iterator gIter = def.geomTags.begin();
  set<int>::const_iterator gLast = def.geomTags.end();
  for (; gIter != gLast; gIter++) {
    for (int iTag = (*gIter); iTag <= (*gIter); iTag++) {
      mob->addGEntity(type,iTag);
    }
  }

  // --- Kinematics ---
  KinematicType kinType = def.kinType;
  if ( kinType == KT_DISPLACEMENT ) {
    mob->setDxKinematics(def.kinExpression);
  }
  else {
    velocityFormulation velType = NOFORMULATION;
    if      (kinType == KT_VELOCITY_TRANSLATION)       velType = TRANSLATION;
    else if (kinType == KT_VELOCITY_FULLRANDOM)        velType = FULLRANDOM;
    else if (kinType == KT_VELOCITY_RANDOMTRANSLATION) velType = RANDOM_TRANSLATION;
    else if (kinType == KT_VELOCITY_ISOMETRY)          velType = ISOMETRY;
    else if (kinType == KT_VELOCITY_ROTATION)          velType = ROTATION;
    else if (kinType == KT_VELOCITY_SHEAR_YZ)          velType = SHEAR_YZ;
    else if (kinType == KT_VELOCITY_SHEAR_ZX)          velType = SHEAR_ZX;
    else if (kinType == KT_VELOCITY_SHEAR_XY)          velType = SHEAR_XY;
    else if (kinType == KT_VELOCITY_EXPRESSION)        velType = PARSED;
    else {
      cerr << "Unknown velocity formulation: "<< kinType <<endl; throw;
    }
    mob->setVKinematics(velType,
                        def.velocity,
                        def.center,
                        def.kinExpression);
  }

  std::list<LocalSFDef>::const_iterator sIter = def.sizes.begin();
  std::list<LocalSFDef>::const_iterator sLast = def.sizes.end();
  for (; sIter != sLast; sIter++) {
    double radius = (*sIter).radius;
    double sizeIn = (*sIter).sizeIn;
    double sizeOut = (*sIter).sizeOut;
    double scale = sizeOut / sizeIn;
    LocalSizeField* locSField = new LocalSizeField(mesh,radius,sizeIn,scale);
    set<int>::const_iterator gIter = def.geomTags.begin();
    set<int>::const_iterator gLast = def.geomTags.end();
    for (; gIter != gLast; gIter++) {
      for (int iTag = (*gIter); iTag <= (*gIter); iTag++) {
        locSField->addGeometricEntity(type,iTag);
      }
    }
    locSField->updateTree();
    mob->addLocalSField(locSField);
  }
}

// ----------------------------------------------------------------------
void deleteObjects(mobileObjectSet* objSet)
{
  set<mobileObject*> objs = objSet->getObjects();
  set<mobileObject*>::iterator oIter = objs.begin();
  for (; oIter != objs.end(); oIter++) {
    
    mobileObject * mob = *oIter;

    // clean size fields
    set<LocalSizeField* > localSFs = mob->getSizes();
    set<LocalSizeField* >::iterator sIter = localSFs.begin();
    for (; sIter != localSFs.end(); sIter++) {
      if (*sIter) delete (*sIter);
    }
    
    // delete object
    if (mob) delete mob;
  }
}

// ----------------------------------------------------------------------
int main(int argc, char* argv[]) 
{
  MoveParameters parameters = getParameters();

  // ------------------------------------------------
  // setup the output
  // ------------------------------------------------

  string outputType = parameters.outType;
  int outputFrequency = parameters.outFrequency;
  string outputPrefix = parameters.outPrefix;

  // ------------------------------------------------
  // load the mesh
  // ------------------------------------------------

  cout << "Loading the mesh...\n";
  double cpu_mesh_0 = CPUTime();

  // --- Reading model ---
  string meshFileName = parameters.meshFileName;
  string geoFileName  = parameters.geoFileName;
  pGModel model = 0;
  GM_create(&model,"theModel");
  if ( !geoFileName.empty() ) GM_readFromGEO(model, geoFileName.c_str());
  else                        GM_readFromMSH(model, meshFileName.c_str());

  // --- Reading mesh ---
  pMesh mesh = M_new(model);
  M_load(mesh,meshFileName.c_str());
//   M_writeMsh (mesh, (outputPrefix + "initMesh.msh").c_str(), 1, NULL);

  double cpu_mesh_tot = CPUTime() - cpu_mesh_0;
  cout << "Loaded the mesh in "<<cpu_mesh_tot<<" seconds\n";

  // ------------------------------------------------
  // build the mesh adapter
  // ------------------------------------------------
  cout << "Buidling the mesh adapter...\n";
  double cpu_ma_0 = CPUTime();
  MeshAdapter* ma = new MeshAdapter(mesh);

  ma->setOutputPrefix(outputPrefix);

  ma->setAlgorithm(CPS_SWP_SLV_SPT);

  int maxInnerIter = parameters.maxInnerIter;
  ma->setMaxIterationsNumber(maxInnerIter);

  double lowerLenghtBound = parameters.lowerLength;
  double upperLenghtBound = parameters.upperLength;
  ma->setEdgeLenSqBounds( lowerLenghtBound * lowerLenghtBound,
                          upperLenghtBound * upperLenghtBound );

  double noSwapQuality = parameters.noSwapQuality;
  ma->setNoSwapQuality(noSwapQuality);
  double swapMinImproveRatio = parameters.swapMinImproveRatio;
  ma->setSwapMinImproveRatio(swapMinImproveRatio);

  double sliverQuality = parameters.sliverQuality;
  ma->setSliverQuality(sliverQuality);

  ma->setCollapseOnBoundary(parameters.collapseOnBoundary,
                            parameters.boundaryTolerance);
  
  ma->setGeoTracking(parameters.trackGeometry);

  double cpu_ma_tot = CPUTime() - cpu_ma_0;
  cout << "Built the mesh adapter in "<<cpu_ma_tot<<" seconds\n";

// #warning "debug"
//   ma->setSliverPermissionInESplit(true);
//   ma->setSliverPermissionInECollapse(true);

  // ------------------------------------------------
  // build the size fields
  // ------------------------------------------------
  cout << "Buidling the size fields...\n";
  double cpu_sf_0 = CPUTime();

  set<pSField> sizeFields;

  std::list<AnalSFDef>::const_iterator sIter = parameters.sizes.begin();
  std::list<AnalSFDef>::const_iterator sLast = parameters.sizes.end();
  for (; sIter != sLast; sIter++) {

    SizeFieldBase* sizeField = NULL;
  
    OrientationType orient = (*sIter).orientation;
    if ( orient == ORT_ISOTROPIC ) {
      sizeField = new AnalyticalSField((*sIter).isoSize);
    }
    else {
      sizeField = new AnalyticalSField((*sIter).anisoSize,
                                       (*sIter).anisoDir0,
                                       (*sIter).anisoDir1,
                                       (*sIter).anisoDir2);
    }

    sizeFields.insert(sizeField);
    ma->addSizeField(sizeField);
  }

  double cpu_sf_tot = CPUTime() - cpu_sf_0;
  cout << "Built the size field in "<<cpu_sf_tot<<" seconds\n";

  ma->printStatistics(cout);

  // ------------------------------------------------
  // setup the quality monitoring
  // ------------------------------------------------

  string qualityFileName = outputPrefix + "quality";
  FILE* qualFile = fopen(qualityFileName.c_str(),"w");
  if (!qualFile) {
    cerr << "Could not open the quality file " << qualityFileName << endl;
    throw;
  }

  double meanShape, worstShape;
  ma->getStatistics(&meanShape,&worstShape);
  fprintf(qualFile,"iter\tmean quality\tworst quality\n");
  fprintf(qualFile,"%d\t%g\t%g\n",0,meanShape,worstShape);
  fclose(qualFile);

  // ------------------------------------------------
  // setup the cpu time monitoring
  // ------------------------------------------------

  string cpuFileName = outputPrefix + "cpu";
  FILE* cpuFile = fopen(cpuFileName.c_str(),"w");
  if (!cpuFile) {
    cerr << "Could not open the cpu report file " << cpuFileName << endl;
    throw;
  }

  fprintf(cpuFile,"iter\tnodes motion\tadaptation\toutputs  \taccumulated nm\taccum adapt\taccum out\ttotal\n");
  fclose(cpuFile);

  string produceCpuInfo = "cat /proc/cpuinfo > \"" + outputPrefix + "cpuinfo\"";
  system(produceCpuInfo.c_str());

  // ------------------------------------------------
  // setup the mesh size monitoring
  // ------------------------------------------------

  string MSFileName = outputPrefix + "meshSize";
  FILE* MSFile = fopen(MSFileName.c_str(),"w");
  if (!MSFile) {
    cerr << "Could not open the mesh size report file " << MSFileName << endl;
    throw;
  }

  fprintf(MSFile,"iter\t#nodes\t#edges\t#faces\t#regions\n");
  fprintf(MSFile,"init\t%d\t%d\t%d\t%d\n",
          M_numVertices(mesh),M_numEdges(mesh),M_numFaces(mesh),M_numRegions(mesh));
  fclose(MSFile);

  // ------------------------------------------------
  // setup debug tools
  // ------------------------------------------------

  bool debugMode = parameters.debugMode;
  ma->setDebugMode(debugMode);

  // --- cross checks between two executions ---
  bool openJournal = parameters.openJournal;
  if ( openJournal ) {
    
    // Will hold a list of operations and checks in this execution
    ma->openJournal();

    // Will compare the list to another one (and abort if different)
    string refJournalName = parameters.referenceJournal;
    if ( !refJournalName.empty() ) {
      std::cout<<"Warning: comparing execution with the journal \'"
               << refJournalName << "\'\n";
      ma->setReferenceJournal(refJournalName);
    }
  }

  // --- slivers output ---
  bool sliverReports = parameters.sliverReports;
  if ( sliverReports ) {
    ma->enableSliverReports();
  }
  bool sliverOps = parameters.testSliverOperators;
  ma->testSliverOperators(sliverOps);

  // ------------------------------------------------
  // TEST MOBILE OBJECTS
  // ------------------------------------------------

  ma->storeInitialCoordinates();

  // ---- build the mobile objects ----
  cout << "Building the mobile objects...\n";
  double cpu_mobj_0 = CPUTime();
  mobileObjectSet* objs = new mobileObjectSet();
  for (unsigned int iObj = 0; iObj < parameters.objects.size(); iObj++) {
    mobileObject* mob = new mobileObject(mesh);
    setMobileObject(mob,mesh,parameters.objects[iObj]);
    objs->insert(mob);
  }
  ma->registerObjects(objs);
  double cpu_mobj_tot = CPUTime() - cpu_mobj_0;
  cout << "Built the mobile objects in "<<cpu_mobj_tot<<" seconds\n";

// #warning "Size field printed to debug"
//   ma->writePos("sizeFieldInit.pos",OD_ISOSIZEFIELD);

  // ---- get time parameters ----
  double dtGlob = parameters.timeStep;
  double finalTime = parameters.finalTime;
  int maxTimeSteps = parameters.maxNbTimeSteps;

  // ---- get node motion parameters ----
  string nodeMotionType = parameters.nodeMotionType;
  double chi = parameters.stiffnessAlteration;

  // ---- the loop ----
  double cpu_t0 = CPUTime();
  double t = 0., dt_iter = 0.;
  int iter = 0;
  double cpu_move = 0., cpu_adapt = 0., cpu_output = 0.;
  while (iter <= maxTimeSteps && t < finalTime)
    {
      double cpu_t1 = CPUTime();
      if ( iter > 0 )  // a first step is performed with the adaptation before any motion occurs
        {
          // --- find time step for this iteration ---
          double t_left_tot = std::max(finalTime - t, 0.0);
          dt_iter = std::min(dtGlob, t_left_tot);
          
          // --- move nodes (dummy move and Laplace smoothing) ---
          if ( !strcmp(nodeMotionType.c_str(),"ForceBoundaries") ) {
            double part;
            if (!ma->partlyMoveObjects(t, dt_iter, &part)) {
              cout <<"Could not perform a forced move\n";
              ma->abort();
            }
            dt_iter *= part;

            if (debugMode) if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort();
            ma->LaplaceSmoothing(OPTIMAL);
            if (debugMode) if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort();
          }

          t += dt_iter;  ma->setTime(t);
          printf("\nGlobal time step %d, t = %f, dt = %f\n\n",iter,t,dt_iter);

          // --- move nodes (elasticity analogy) ---
          if ( !strcmp(nodeMotionType.c_str(),"ElasticAnalogy") ) {
            if (!ma->moveObjectsAndReposition(t, dt_iter, chi)) {
              cout <<"Could not perform a forced move\n";
              ma->abort();
            }
          }
          if (debugMode) if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort();
        }
      double cpu_t2 = CPUTime(); 
      cpu_move += ( cpu_t2 - cpu_t1 );

      // --- adapt the mesh ---
      if (debugMode) if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort();
      ma->run();
      if (debugMode) if ( !checkMesh(mesh,CHECK_ALL,1) )  ma->abort();
      double cpu_t3 = CPUTime();
      cpu_adapt += ( cpu_t3 - cpu_t2 );

      // --- outputs ---

      {  // full mesh
        if ( (iter%outputFrequency) == 0 ) {
          writeSolution(ma,mesh,iter,outputType);
        }
      }

      {  // quality monitor
        ma->getStatistics(&meanShape,&worstShape);
        FILE* qf = fopen(qualityFileName.c_str(),"a");
        fprintf(qf,"%d\t%e\t%e\n",iter,meanShape,worstShape);
        fclose(qf);
      }

      {  // cpu monitor
        double cpu_t4 = CPUTime();
        cpu_output += ( cpu_t4 - cpu_t3 );
      
        FILE* cpuF = fopen(cpuFileName.c_str(),"a");
        fprintf(cpuF,"%d\t%e\t%e\t%e\t%e\t%e\t%e\t%e\n",
                iter, cpu_t2 - cpu_t1, cpu_t3 - cpu_t2, cpu_t4 - cpu_t3,
                cpu_move, cpu_adapt, cpu_output, cpu_t4 - cpu_t0 );
        fclose(cpuF);
      }

      {  // mesh size monitor
        FILE* msf = fopen(MSFileName.c_str(),"a");
        fprintf(msf,"%d\t%d\t%d\t%d\t%d\n",
                iter,M_numVertices(mesh),M_numEdges(mesh),M_numFaces(mesh),M_numRegions(mesh));
        fclose(msf);
      }

      { // slivers stats
        if ( (iter%10) == 0 ) {
          ofstream sliverOut( (outputPrefix + "slivers_tmp").c_str() );
          ma->printSliverRegionStatistics(sliverOut);
        }
      }

      { // operations stats
        if ( (iter%10) == 0 ) {
          ofstream operationsOut( (outputPrefix + "statistics_tmp").c_str() );
          ma->printStatistics(operationsOut);
        }
      }

      { // journal
        if ( (iter%10) == 0 ) {
          ofstream journalOut( (outputPrefix + "journal_tmp").c_str() );
          ma->flushJournal(journalOut);
        }
      }

      iter++;
    }

  if (objs) {
    deleteObjects(objs);
    delete objs;
  }

  ma->removeStoredCoordinates();

  // ------------------------------------------------
  // Write final result
  // ------------------------------------------------

  ma->writeMsh("result.msh");
  ma->writePos("result.pos",OD_MEANRATIO);

  ma->printStatistics(cout);
  ofstream statOut((outputPrefix + "statistics").c_str());
  ma->printStatistics(statOut);

  ma->printSliverRegionStatistics(cout); 
  ofstream sliverOut((outputPrefix + "slivers").c_str());
  ma->printSliverRegionStatistics(sliverOut);

  ofstream journalOut((outputPrefix + "journal").c_str());
  ma->flushJournal(journalOut);

  // ------------------------------------------------
  // Clean up
  // ------------------------------------------------

  if (ma) delete ma;
  set<pSField>::iterator sfIter = sizeFields.begin();
  set<pSField>::iterator sfLast = sizeFields.end();
  for (; sfIter != sfLast; sfIter++) {
    delete (*sfIter);
  }
  
  if (mesh)  M_delete(mesh);
  if (model) GM_delete(model);
}

// ----------------------------------------------------------------------
