// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: Jean-Francois Remacle, Gaetan Compere
// -------------------------------------------------------------------

#include <signal.h>
#ifndef _WIN_
#include <sys/time.h>
#include <sys/resource.h>
#else
#include <windows.h>
#include <psapi.h>
#endif
#include <stdio.h>
#include <stdarg.h>
#include "MeshDataBaseMessage.h"
using std::string;

namespace MAd {

  // Handle signals. We should not use Msg functions in these...

  void Signal(int sig_num)
  {
    switch (sig_num) {
    case SIGSEGV:
      Msg(MDB_FATAL, "Segmentation violation (invalid memory reference)");
      break;
    case SIGFPE:
      Msg(MDB_FATAL, "Floating point exception (division by zero?)");
      break;
    case SIGINT:
      Msg(MDB_FATAL, "Interrupt (generated from terminal special char)");
      break;
    default:
      Msg(MDB_FATAL, "Unknown signal");
      break;
    }
  }

  // General purpose message routine

  void Msg(int level, string fmt, ...)
  {
    va_list args;
    va_start(args, fmt);

    Msg_char(level, fmt.c_str(), args);

    va_end(args);
  }

  void Msg_char(int level, const char *fmt, ...)
  {
    va_list args;
    int abort = 0;

    va_start(args, fmt);

    switch (level) {

    case MDB_PROGRESS:
    case MDB_STATUS1N:
    case MDB_STATUS2N:
    case MDB_STATUS3N:
      break;

    case MDB_DIRECT:
      vfprintf(stdout, fmt, args);
      fprintf(stdout, "\n");
      break;

    case MDB_FATAL:
    case MDB_FATAL3: abort = 1;
    case MDB_FATAL1:
    case MDB_FATAL2:
      fprintf(stderr, MDB_FATAL_STR);
      vfprintf(stderr, fmt, args);
      fprintf(stderr, "\n");
      break;

    case MDB_GERROR:
    case MDB_GERROR1:
    case MDB_GERROR2:
    case MDB_GERROR3:
      fprintf(stderr, MDB_ERROR_STR);
      vfprintf(stderr, fmt, args);
      fprintf(stderr, "\n");
      break;

    case MDB_WARNING:
    case MDB_WARNING1:
    case MDB_WARNING2:
    case MDB_WARNING3:
      fprintf(stderr, MDB_WARNING_STR);
      vfprintf(stderr, fmt, args);
      fprintf(stderr, "\n");
      break;

      //   case DEBUG:
      //   case DEBUG1:
      //   case DEBUG2:
      //   case DEBUG3:
      //     fprintf(stderr, DEBUG_STR);
      //     vfprintf(stderr, fmt, args);
      //     fprintf(stderr, "\n");
      //     break;
    
    default:
      fprintf(stderr, MDB_INFO_STR);
      vfprintf(stderr, fmt, args);
      fprintf(stderr, "\n");
      break;
    }

    va_end(args);

    if(abort)
      throw;
  }

  // CPU time computation

  void GetResources(long *s, long *us, long *mem)
  {
#ifndef _WIN_
    static struct rusage r;

    getrusage(RUSAGE_SELF, &r);
    *s = (long)r.ru_utime.tv_sec;
    *us = (long)r.ru_utime.tv_usec;
    *mem = (long)r.ru_maxrss;
#else
	PROCESS_MEMORY_COUNTERS counters;
    if (GetProcessMemoryInfo (GetCurrentProcess(), &counters, sizeof (counters)))
        *mem=counters.PagefileUsage;
	SYSTEMTIME systime; 
	GetSystemTime(&systime); 
	*s=systime.wSecond;
	*us=systime.wMilliseconds;
#endif
  }
}
